@@ -9,20 +9,22 @@ import asyncio
 from collections.abc import Callable, MutableMapping
 import logging
 from statistics import fmean
-from typing import Any, Final, Literal
+from types import TracebackType
+from typing import Any, Final, Literal, Self
 
 from bleak import BleakClient
 from bleak.backends.characteristic import BleakGATTCharacteristic
 from bleak.backends.device import BLEDevice
-from bleak.exc import BleakError
+from bleak.exc import BleakCharacteristicNotFoundError, BleakError
 from bleak_retry_connector import BLEAK_TIMEOUT, establish_connection
 
-from aiobmsble import BMSdp, BMSsample, BMSvalue, MatcherPattern
+from aiobmsble import BMSDp, BMSInfo, BMSSample, BMSValue, MatcherPattern
 
 
 class BaseBMS(ABC):
     """Abstract base class for battery management system."""
 
+    INFO: BMSInfo  # static BMS info, set "default_" keys in subclass
     MAX_RETRY: Final[int] = 3  # max number of retries for data requests
     TIMEOUT: Final[float] = BLEAK_TIMEOUT / 4  # default timeout for BMS operations
     # calculate time between retries to complete all retries (2 modes) in TIMEOUT seconds
@@ -44,26 +46,32 @@ class BaseBMS(ABC):
     def __init__(
         self,
         ble_device: BLEDevice,
-        reconnect: bool = False,
+        keep_alive: bool = True,
         logger_name: str = "",
     ) -> None:
         """Intialize the BMS.
 
-        notification_handler: the callback function used for notifications from 'uuid_rx()'
+        `_notification_handler`: the callback function used for notifications from `uuid_rx()`
             characteristic. Not defined as abstract in this base class, as it can be both,
             a normal or async function
 
         Args:
-            logger_name (str): name of the logger for the BMS instance (usually file name)
             ble_device (BLEDevice): the Bleak device to connect to
-            reconnect (bool): if true, the connection will be closed after each update
+            bms_info (dict[Literal["manufacturer", "model"], str]): default BMS identification
+            keep_alive (bool): if true, the connection will be kept active after each update.
+                Make sure to call `disconnect()` when done using the BMS class or better use
+                `async with` context manager (requires `keep_alive=True`).
+            logger_name (str): name of the logger for the BMS instance, default: module name
 
         """
         assert (
             getattr(self, "_notification_handler", None) is not None
-        ), "BMS class must define _notification_handler method"
+        ), "BMS class must define `_notification_handler` method"
+        assert {"default_manufacturer", "default_model"}.issubset(
+            self.INFO
+        ), "BMS class must define `INFO`"
         self._ble_device: Final[BLEDevice] = ble_device
-        self._reconnect: Final[bool] = reconnect
+        self._keep_alive: Final[bool] = keep_alive
         self.name: Final[str] = self._ble_device.name or "undefined"
         self._inv_wr_mode: bool | None = None  # invert write mode (WNR <-> W)
         logger_name = logger_name or self.__class__.__module__
@@ -73,7 +81,7 @@ class BaseBMS(ABC):
         )
 
         self._log.debug(
-            "initializing %s, BT address: %s", self.device_id(), ble_device.address
+            "initializing %s, BT address: %s", self.bms_id(), ble_device.address
         )
         self._client: BleakClient = BleakClient(
             self._ble_device,
@@ -83,6 +91,22 @@ class BaseBMS(ABC):
         self._data: bytearray = bytearray()
         self._data_event: Final[asyncio.Event] = asyncio.Event()
 
+    async def __aenter__(self) -> Self:
+        """Asynchronous context manager to implement `async with` functionality."""
+        if not self._keep_alive:
+            raise ValueError("usage of context manager requires `keep_alive=True`.")
+        await self._connect()
+        return self
+
+    async def __aexit__(
+        self,
+        typ: type[BaseException] | None,
+        exc: BaseException | None,
+        tb: TracebackType | None,
+    ) -> None:
+        """Asynchronous context manager exit functionality."""
+        await self.disconnect()
+
     @classmethod
     def get_bms_module(cls) -> str:
         """Return BMS module name, e.g. aiobmsble.bms.dummy_bms."""
@@ -93,18 +117,10 @@ class BaseBMS(ABC):
     def matcher_dict_list() -> list[MatcherPattern]:
         """Return a list of Bluetooth advertisement matchers."""
 
-    @staticmethod
-    @abstractmethod
-    def device_info() -> dict[str, str]:
-        """Return a dictionary of device information.
-
-        keys: manufacturer, model
-        """
-
     @classmethod
-    def device_id(cls) -> str:
-        """Return device information as string."""
-        return " ".join(cls.device_info().values())
+    def bms_id(cls) -> str:
+        """Return static BMS information as string."""
+        return f"{cls.INFO['default_manufacturer']} {cls.INFO['default_model']}"
 
     @staticmethod
     @abstractmethod
@@ -121,8 +137,58 @@ class BaseBMS(ABC):
     def uuid_tx() -> str:
         """Return 16-bit UUID of characteristic that provides write property."""
 
+    async def device_info(self) -> BMSInfo:
+        """Return a dictionary of device information.
+
+        keys: manufacturer, model, model_id, name, serial_number, sw_version, hw_version
+        """
+
+        disconnect: Final[bool] = not self._client.is_connected
+        await self._connect()
+        dev_info: Final[BMSInfo] = await self._fetch_device_info()
+        if disconnect:
+            await self.disconnect()
+
+        self._log.debug("BMS info %s", dev_info)
+        return dev_info
+
+    async def _fetch_device_info(self) -> BMSInfo:
+        """Fetch the device information via BLE."""
+        if not self._client.services.get_service("180a"):
+            self._log.debug("No BT device information available.")
+            return BMSInfo()
+
+        type CharacteristicType = Literal[
+            "model",
+            "serial_number",
+            "fw_version",
+            "sw_version",
+            "hw_version",
+            "manufacturer",
+        ]
+
+        characteristics: Final[tuple[tuple[str, CharacteristicType], ...]] = (
+            ("2a24", "model"),
+            ("2a25", "serial_number"),
+            ("2a26", "fw_version"),
+            ("2a27", "hw_version"),
+            ("2a28", "sw_version"),  # overwrite FW with SW version
+            ("2a29", "manufacturer"),
+        )
+
+        info: BMSInfo = BMSInfo()
+        for char, key in characteristics:
+            try:
+                if value := await self._client.read_gatt_char(char):
+                    info[key] = barr2str(value)
+                    self._log.debug("BT device %s: '%s'", key, info[key])
+            except BleakCharacteristicNotFoundError:
+                pass
+
+        return info
+
     @staticmethod
-    def _calc_values() -> frozenset[BMSvalue]:
+    def _calc_values() -> frozenset[BMSValue]:
         """Return values that the BMS cannot provide and need to be calculated.
 
         See _add_missing_values() function for the required input to actually do so.
@@ -130,7 +196,7 @@ class BaseBMS(ABC):
         return frozenset()
 
     @staticmethod
-    def _add_missing_values(data: BMSsample, values: frozenset[BMSvalue]) -> None:
+    def _add_missing_values(data: BMSSample, values: frozenset[BMSValue]) -> None:
         """Calculate missing BMS values from existing ones.
 
         Args:
@@ -144,7 +210,7 @@ class BaseBMS(ABC):
         if not values or not data:
             return
 
-        def can_calc(value: BMSvalue, using: frozenset[BMSvalue]) -> bool:
+        def can_calc(value: BMSValue, using: frozenset[BMSValue]) -> bool:
             """Check value to add does not exist, is requested, and needed data is available."""
             return (value in values) and (value not in data) and using.issubset(data)
 
@@ -152,7 +218,7 @@ class BaseBMS(ABC):
         battery_level: Final[int | float] = data.get("battery_level", 0)
         current: Final[float] = data.get("current", 0)
 
-        calculations: dict[BMSvalue, tuple[set[BMSvalue], Callable[[], Any]]] = {
+        calculations: dict[BMSValue, tuple[set[BMSValue], Callable[[], Any]]] = {
             "voltage": ({"cell_voltages"}, lambda: round(sum(cell_voltages), 3)),
             "delta_voltage": (
                 {"cell_voltages"},
@@ -177,6 +243,10 @@ class BaseBMS(ABC):
                 {"voltage", "cycle_charge"},
                 lambda: round(data.get("voltage", 0) * data.get("cycle_charge", 0), 3),
             ),
+            "cycles": (
+                {"design_capacity", "total_charge"},
+                lambda: data.get("total_charge", 0) // data.get("design_capacity", 0),
+            ),
             "power": (
                 {"voltage", "current"},
                 lambda: round(data.get("voltage", 0) * current, 3),
@@ -356,8 +426,8 @@ class BaseBMS(ABC):
             if reset:
                 self._inv_wr_mode = None  # reset write mode
             await self._client.disconnect()
-        except BleakError:
-            self._log.warning("disconnect failed!")
+        except (BleakError, TimeoutError) as exc:
+            self._log.error("disconnect failed! (%s)", type(exc).__name__)
 
     async def _wait_event(self) -> None:
         """Wait for data event and clear it."""
@@ -365,10 +435,10 @@ class BaseBMS(ABC):
         self._data_event.clear()
 
     @abstractmethod
-    async def _async_update(self) -> BMSsample:
+    async def _async_update(self) -> BMSSample:
         """Return a dictionary of BMS values (keys need to come from the SENSOR_TYPES list)."""
 
-    async def async_update(self) -> BMSsample:
+    async def async_update(self, raw: bool = False) -> BMSSample:
         """Retrieve updated values from the BMS using method of the subclass.
 
         Args:
@@ -376,15 +446,16 @@ class BaseBMS(ABC):
                 any calculations or missing values added
 
         Returns:
-            BMSsample: dictionary with BMS values
+            BMSSample: dictionary with BMS values
 
         """
         await self._connect()
 
-        data: BMSsample = await self._async_update()
-        self._add_missing_values(data, self._calc_values())
+        data: BMSSample = await self._async_update()
+        if not raw:
+            self._add_missing_values(data, self._calc_values())
 
-        if self._reconnect:
+        if not self._keep_alive:
             # disconnect after data update to force reconnect next time (slow!)
             await self.disconnect()
 
@@ -392,13 +463,13 @@ class BaseBMS(ABC):
 
     @staticmethod
     def _decode_data(
-        fields: tuple[BMSdp, ...],
+        fields: tuple[BMSDp, ...],
         data: bytearray | dict[int, bytearray],
         *,
         byteorder: Literal["little", "big"] = "big",
         offset: int = 0,
-    ) -> BMSsample:
-        result: BMSsample = {}
+    ) -> BMSSample:
+        result: BMSSample = {}
         for field in fields:
             if isinstance(data, dict) and field.idx not in data:
                 continue
@@ -478,20 +549,31 @@ class BaseBMS(ABC):
 
         """
         return [
-            value / divider if divider != 1 else value
+            (value - offset) / divider
             for idx in range(values)
             if (len(data) >= start + (idx + 1) * size)
             and (
-                value := int.from_bytes(
-                    data[start + idx * size : start + (idx + 1) * size],
-                    byteorder=byteorder,
-                    signed=signed,
+                (
+                    value := int.from_bytes(
+                        data[start + idx * size : start + (idx + 1) * size],
+                        byteorder=byteorder,
+                        signed=signed,
+                    )
                 )
-                - offset
+                or (offset == 0)
             )
         ]
 
 
+def barr2str(barr: bytearray) -> str:
+    """Decode a bytearray to string, stopping at the first non-printable character."""
+    s = barr.decode("utf-8", errors="ignore")
+    for i, c in enumerate(s):
+        if not c.isprintable():
+            return s[:i].strip()
+    return s.strip()
+
+
 def crc_modbus(data: bytearray) -> int:
     """Calculate CRC-16-CCITT MODBUS."""
     crc: int = 0xFFFF
