@@ -8,16 +8,24 @@ from uuid import UUID
 from bleak.assigned_numbers import CharacteristicPropertyName
 from bleak.backends.characteristic import BleakGATTCharacteristic
 from bleak.backends.device import BLEDevice
+from bleak.backends.service import BleakGATTServiceCollection
 from bleak.exc import BleakError
 from bleak.uuids import normalize_uuid_str
 import pytest
 
-from aiobmsble import BMSdp, BMSsample, MatcherPattern
-from aiobmsble.basebms import BaseBMS, crc8, crc_modbus, crc_sum, crc_xmodem, lrc_modbus
+from aiobmsble import BMSDp, BMSInfo, BMSSample, BMSValue, MatcherPattern
+from aiobmsble.basebms import (
+    BaseBMS,
+    barr2str,
+    crc8,
+    crc_modbus,
+    crc_sum,
+    crc_xmodem,
+    lrc_modbus,
+)
 from aiobmsble.bms.dummy_bms import BMS as DummyBMS
-
-from .bluetooth import generate_ble_device
-from .conftest import MockBleakClient
+from tests.bluetooth import generate_ble_device
+from tests.conftest import MockBleakClient
 
 
 class MockWriteModeBleakClient(MockBleakClient):
@@ -67,16 +75,16 @@ class MockWriteModeBleakClient(MockBleakClient):
 class MinTestBMS(BaseBMS):
     """Minimal Test BMS implementation."""
 
+    INFO: BMSInfo = {
+        "default_manufacturer": "Test Manufacturer",
+        "default_model": "minimal BMS for test",
+    }
+
     @staticmethod
     def matcher_dict_list() -> list[MatcherPattern]:
         """Provide BluetoothMatcher definition."""
         return [{"local_name": "Test", "connectable": True}]
 
-    @staticmethod
-    def device_info() -> dict[str, str]:
-        """Return device information for the battery management system."""
-        return {"manufacturer": "Test Manufacturer", "model": "minimal BMS for test"}
-
     @staticmethod
     def uuid_services() -> list[str]:
         """Return list of 128-bit UUIDs of services required by BMS."""
@@ -99,23 +107,42 @@ class MinTestBMS(BaseBMS):
         self._log.debug("RX BLE data: %s", data)
         # do not set event to make tests fail if wait_for_notify is not set
 
-    async def _async_update(self) -> BMSsample:
+    async def _async_update(self) -> BMSSample:
         """Update battery status information."""
         await self._await_reply(b"mock_command", wait_for_notify=False)  # do not wait
         return {"problem_code": 21}
 
 
+class DataTestBMS(MinTestBMS):
+    """BMS providing simple data to test, e.g. value calculation."""
+
+    @staticmethod
+    def _calc_values() -> frozenset[BMSValue]:
+        return frozenset({"power", "battery_charging"})
+
+    async def _async_update(self) -> BMSSample:
+        """Update battery status information."""
+        await self._await_reply(b"mock_command", wait_for_notify=False)  # do not wait
+        return {
+            "voltage": 13,
+            "current": 1.7,
+            "cycle_charge": 19,
+            "cycles": 23,
+            "problem_code": 21,
+        }
+
+
 class WMTestBMS(MinTestBMS):
-    """Test BMS implementation."""
+    """Write mode mock BMS implementation."""
 
     def __init__(
         self,
         char_tx_properties: list[str],
         ble_device: BLEDevice,
-        reconnect: bool = False,
+        keep_alive: bool = True,
     ) -> None:
         """Initialize BMS."""
-        super().__init__(ble_device, reconnect)
+        super().__init__(ble_device, keep_alive)
         self._char_tx_properties: list[str] = char_tx_properties
 
     def _wr_response(self, char: int | str) -> bool:
@@ -129,17 +156,66 @@ class WMTestBMS(MinTestBMS):
         self._data = data
         self._data_event.set()
 
-    async def _async_update(self) -> BMSsample:
+    async def _async_update(self) -> BMSSample:
         """Update battery status information."""
         await self._await_reply(b"mock_command")
 
         return {"problem_code": int.from_bytes(self._data, "big", signed=False)}
 
 
-def test_calc_missing_values(bms_data_fixture: BMSsample) -> None:
+@pytest.mark.parametrize(
+    ("bt_patch", "result_patch"),
+    [
+        ({"2a28": b"mock_SW_version"}, {"sw_version": "mock_SW_version"}),
+        ({"2a28": b""}, {}),
+        ({}, {}),
+    ],
+    ids=["defaults", "empty", "no_char"],
+)
+async def test_device_info(
+    monkeypatch: pytest.MonkeyPatch,
+    patch_bleak_client: Callable[..., None],
+    bt_patch,
+    result_patch,
+) -> None:
+    """Verify that device_info reads BLE characteristic 180A and provides default values."""
+    defaults = MockBleakClient.BT_INFO.copy()
+    del defaults["2a28"]
+    monkeypatch.setattr(MockBleakClient, "BT_INFO", defaults | bt_patch)
+    patch_bleak_client()
+    bms: MinTestBMS = MinTestBMS(generate_ble_device())
+    assert (
+        await bms.device_info()
+        == {
+            "fw_version": "mock_FW_version",
+            "model": "mock_model",
+            "serial_number": "mock_serial_number",
+            "hw_version": "mock_HW_version",
+            "manufacturer": "mock_manufacturer",
+        }
+        | result_patch
+    )
+
+
+async def test_device_info_fail(
+    monkeypatch: pytest.MonkeyPatch, patch_bleak_client: Callable[..., None]
+) -> None:
+    """Test only BMS default information is returned if characteristic 0x180a does not exit."""
+    monkeypatch.setattr(
+        BleakGATTServiceCollection, "get_service", lambda obj, char: None
+    )
+    patch_bleak_client()
+    bms: MinTestBMS = MinTestBMS(generate_ble_device())
+    await bms.async_update()  # run update to have connection open
+    assert not await bms.device_info()  # if characteristic does not exist, no ouput
+    assert bms.name == "MockBLEDevice"  # name is gathered from BLEDevice
+    assert bms._client.is_connected
+
+
+def test_calc_missing_values(bms_data_fixture: BMSSample) -> None:
     """Check if missing data is correctly calculated."""
-    bms_data: BMSsample = bms_data_fixture
-    ref: BMSsample = bms_data_fixture.copy()
+    bms_data: BMSSample = bms_data_fixture
+    ref: BMSSample = bms_data_fixture.copy()
 
     BaseBMS._add_missing_values(
         bms_data,
@@ -151,7 +227,7 @@ def test_calc_missing_values(bms_data_fixture: BMSsample) -> None:
                 "runtime",
                 "delta_voltage",
                 "temperature",
-                "voltage",  # check that not overwritten
+                "voltage",  # check value is not overwritten
             }
         ),
     )
@@ -176,28 +252,58 @@ def test_calc_missing_values(bms_data_fixture: BMSsample) -> None:
 
 def test_calc_voltage() -> None:
     """Check if missing data is correctly calculated."""
-    bms_data: BMSsample = {"cell_voltages": [3.456, 3.567]}
-    ref: BMSsample = bms_data.copy()
+    bms_data: BMSSample = {"cell_voltages": [3.456, 3.567]}
+    ref: BMSSample = bms_data.copy()
     BaseBMS._add_missing_values(bms_data, frozenset({"voltage"}))
     assert bms_data == ref | {"voltage": 7.023, "problem": False}
 
 
 def test_calc_cycle_chrg() -> None:
     """Check if missing data is correctly calculated."""
-    bms_data: BMSsample = {"battery_level": 73, "design_capacity": 125}
-    ref: BMSsample = bms_data.copy()
+    bms_data: BMSSample = {"battery_level": 73, "design_capacity": 125}
+    ref: BMSSample = bms_data.copy()
     BaseBMS._add_missing_values(bms_data, frozenset({"cycle_charge"}))
     assert bms_data == ref | {"cycle_charge": 91.25, "problem": False}
 
 
 def test_calc_battery_level() -> None:
     """Check if missing battery_level is correctly calculated."""
-    bms_data: BMSsample = {"cycle_charge": 421, "design_capacity": 983}
-    ref: BMSsample = bms_data.copy()
+    bms_data: BMSSample = {"cycle_charge": 421, "design_capacity": 983}
+    ref: BMSSample = bms_data.copy()
     BaseBMS._add_missing_values(bms_data, frozenset({"battery_level"}))
     assert bms_data == ref | {"battery_level": 42.8, "problem": False}
 
 
+def test_calc_cycles() -> None:
+    """Check if missing cycle is correctly calculated."""
+    bms_data: BMSSample = {"total_charge": 1234567, "design_capacity": 256}
+    ref: BMSSample = bms_data.copy()
+    BaseBMS._add_missing_values(bms_data, frozenset({"cycles"}))
+    assert bms_data == ref | {"cycles": 4822, "problem": False}
+
+
+@pytest.mark.parametrize(
+    ("raw", "expected"),
+    [(True, {}), (False, {"battery_charging": True, "power": 22.1, "problem": True})],
+    ids=["raw", "ext"],
+)
+async def test_async_update(
+    patch_bleak_client: Callable[..., None], raw: bool, expected: dict
+) -> None:
+    """Check update function of the BMS returns calues."""
+    patch_bleak_client()
+    bms: DataTestBMS = DataTestBMS(generate_ble_device())
+    assert await bms.async_update(raw=raw) == BMSSample(
+        {
+            "voltage": 13,
+            "current": 1.7,
+            "cycle_charge": 19,
+            "cycles": 23,
+            "problem_code": 21,
+        }
+    ) | BMSSample(**expected)
+
+
 @pytest.mark.parametrize(
     ("problem_sample"),
     [
@@ -212,9 +318,9 @@ def test_calc_battery_level() -> None:
     ],
     ids=lambda param: param[1],
 )
-def test_problems(problem_sample: tuple[BMSsample, str]) -> None:
+def test_problems(problem_sample: tuple[BMSSample, str]) -> None:
     """Check if missing data is correctly calculated."""
-    bms_data: BMSsample = problem_sample[0].copy()
+    bms_data: BMSSample = problem_sample[0].copy()
 
     BaseBMS._add_missing_values(bms_data, frozenset({"runtime"}))
 
@@ -328,9 +434,9 @@ async def test_no_notify(
     """Test BMS update without waiting for notification event."""
     patch_bleak_client(MockBleakClient)
 
-    bms: MinTestBMS = MinTestBMS(generate_ble_device(), reconnect=True)
+    bms: MinTestBMS = MinTestBMS(generate_ble_device(), keep_alive=False)
     with caplog.at_level(DEBUG):
-        result: BMSsample = await bms.async_update()
+        result: BMSSample = await bms.async_update()
     assert "MockBleakClient write_gatt_char afe2, data: b'mock_command'" in caplog.text
     assert result == {"problem_code": 21}
     assert not bms._client.is_connected
@@ -349,9 +455,9 @@ async def test_disconnect_fail(
     monkeypatch.setattr(MockBleakClient, "disconnect", _raise_bleak_error)
     patch_bleak_client(MockBleakClient)
 
-    bms: MinTestBMS = MinTestBMS(generate_ble_device(), reconnect=True)
+    bms: MinTestBMS = MinTestBMS(generate_ble_device(), keep_alive=False)
     with caplog.at_level(DEBUG):
-        result: BMSsample = await bms.async_update()
+        result: BMSSample = await bms.async_update()
     assert result == {"problem_code": 21}
     assert "failed to disconnect stale connection (BleakError)" in caplog.text
     assert "disconnect failed!" in caplog.text
@@ -375,6 +481,37 @@ async def test_init_connect_fail(
         await bms.async_update()
 
 
+async def test_context_mgr(
+    patch_bleak_client: Callable[..., None],
+) -> None:
+    """Test that context manager provides data."""
+    patch_bleak_client(MockBleakClient)
+
+    async with DataTestBMS(generate_ble_device(), keep_alive=True) as bms:
+        assert await bms.async_update() == {
+            "voltage": 13,
+            "current": 1.7,
+            "battery_charging": True,
+            "power": 22.1,
+            "cycle_charge": 19,
+            "cycles": 23,
+            "problem": True,
+            "problem_code": 21,
+        }
+
+
+async def test_context_mgr_fail(
+    patch_bleak_client: Callable[..., None],
+) -> None:
+    """Check that context manager enforces `keep_alive=True`."""
+
+    patch_bleak_client(MockBleakClient)
+
+    with pytest.raises(ValueError, match="usage of context manager*"):
+        async with MinTestBMS(generate_ble_device(), keep_alive=False) as bms:
+            await bms.async_update()
+
+
 def test_crc_calculations() -> None:
     """Check if CRC calculations are correct."""
     # Example data for CRC calculation
@@ -487,6 +624,32 @@ def test_cell_voltages(data, cells, start, size, byteorder, divider, expected) -
         (bytearray([0x00, 0x7D]), 0, 0, 2, "big", True, 0, 1, []),
         # Divider = 1, offset = 7
         (bytearray([0x00, 0x14]), 1, 0, 2, "big", True, 7, 1, [13]),
+        # no offset, div = 10
+        (
+            bytearray(b"\x65\x64\x00\x40"),
+            4,
+            0,
+            1,
+            "big",
+            True,
+            0,
+            10,
+            [10.1, 10.0, 0.0, 6.4],
+        ),
+        # offset -40, div = 10
+        (
+            bytearray(b"\x65\x64\x00\x40"),
+            4,
+            0,
+            1,
+            "big",
+            True,
+            40,
+            10,
+            [6.1, 6.0, 2.4],
+        ),
+        # offset -25, div = 1
+        (bytearray(b"\x65\x64\x00\x40"), 4, 0, 1, "big", True, 25, 1, [76, 75, 39]),
     ],
     ids=[
         "two_signed_big_endian",
@@ -496,6 +659,9 @@ def test_cell_voltages(data, cells, start, size, byteorder, divider, expected) -
         "not_enough_data",
         "zero_values",
         "divider1_offset10",
+        "no offset",
+        "offset-40",
+        "div-1",
     ],
 )
 def test_temp_values(
@@ -521,8 +687,8 @@ def test_temp_values(
         # Test with big endian and multiple data points
         (
             (
-                BMSdp("voltage", 0, size=2, signed=False),
-                BMSdp("current", 2, size=2, signed=True),
+                BMSDp("voltage", 0, size=2, signed=False),
+                BMSDp("current", 2, size=2, signed=True),
             ),
             bytearray([0x0D, 0x80, 0xFF, 0xEC]),
             "big",
@@ -531,7 +697,7 @@ def test_temp_values(
         ),
         # Test with dict data, little endian
         (
-            (BMSdp("voltage", 0, size=2, signed=False, fct=lambda x: x / 10, idx=1),),
+            (BMSDp("voltage", 0, size=2, signed=False, fct=lambda x: x / 10, idx=1),),
             {1: bytearray([0x64, 0x00])},
             "little",
             0,
@@ -539,7 +705,7 @@ def test_temp_values(
         ),
         # Test with missing dict data
         (
-            (BMSdp("voltage", 0, size=2, signed=False, idx=2),),
+            (BMSDp("voltage", 0, size=2, signed=False, idx=2),),
             {1: bytearray([0x64, 0x00])},
             "little",
             0,
@@ -547,7 +713,7 @@ def test_temp_values(
         ),
         # Test with offset
         (
-            (BMSdp("battery_level", 1, size=1, signed=False),),
+            (BMSDp("battery_level", 1, size=1, signed=False),),
             bytearray([0x00, 0x7D]),
             "big",
             0,
@@ -555,7 +721,7 @@ def test_temp_values(
         ),
         # Test with offset shifting the slice
         (
-            (BMSdp("voltage", 0, size=2, signed=False),),
+            (BMSDp("voltage", 0, size=2, signed=False),),
             bytearray([0x00, 0x00, 0x12, 0x34]),
             "big",
             2,
@@ -572,7 +738,17 @@ def test_temp_values(
 )
 def test_decode_data(fields, data, byteorder, offset, expected) -> None:
     """Test the _decode_data method of BaseBMS with various input parameters."""
-    result: BMSsample = BaseBMS._decode_data(
+    result: BMSSample = BaseBMS._decode_data(
         fields, data, byteorder=byteorder, offset=offset
     )
     assert result == expected
+
+
+@pytest.mark.parametrize(
+    ("data", "expected"),
+    [(b"", ""), (b"\x00 ", ""), (b"test\x00 ", "test"), (b"test  \t\r ", "test")],
+    ids=["empty", "hex", "text_hex", "test_space"],
+)
+def test_barr2str(data: bytes, expected: str) -> None:
+    """Test bytearray to string conversion function."""
+    assert barr2str(bytearray(data)) == expected
